/* life3d.c
 * Nadim Abdo 1998
 *
 * Simulation of life running in 3D
 * For Operating Systems 304-427B Assignment #3
 */

/* =================================================== */
/* ==================== Includes ===================== */
#include <stdio.h>
#include <stdlib.h>
#include <sys/ipc.h>
#include <sys/shm.h>  
#include <sys/sem.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/msg.h>

#include "semaph.h" /*Steven's Unix Net Programming sem utils */
#include "life3d.h"
#include "unistd.h"
#include "face.h"
#include "life3dviewer.h"


/* =================================================== */
/* =============== Function Prototypes =============== */
void dbg_printsem(int sem);

/* =================================================== */
/* ============== InitLifeIPCResources() ============= */
void InitLifeIPCResources()
{
  /* create and init the watchdog semaphore*/
  if (-1==(sem_cleanup_watchdog=semget(IPC_PRIVATE,1,SEM_R | SEM_A)))
    {
      perror("Could not create semaphore watc\n");
      exit(1);
    }
  
  /* signal the semaphore to mark that one more process is using
     the IPC resources*/
  semsignalundo(sem_cleanup_watchdog);
  
  
  /*create and init the shared mem protection semaphore*/
  
  if(-1==(sem_shared_mem=semget(IPC_PRIVATE,1,SEM_R | SEM_A)))
    {
      perror("Could not create semaphore\n");
      exit(1);
    }

  /*Init to +1*/
  semsignal(sem_shared_mem);
  

  /*time synchronization semaphores*/
  if(-1==(sem_time_slice_start=semget(IPC_PRIVATE,1,SEM_R | SEM_A)))
    {
      perror("Could not create semaphore\n");
      exit(1);
    }
  
  if(-1==(sem_time_slice_end=semget(IPC_PRIVATE,1,SEM_R | SEM_A)))
    {
      perror("Could not create semaphore\n");
      exit(1);
    }
  if(-1==(sem_num_updated_faces=semget(IPC_PRIVATE,1,SEM_R | SEM_A)))
    {
      perror("Could not create semaphore\n");
      exit(1);
    }
  if(-1==(sem_faces_commited_next_state=semget(IPC_PRIVATE,1,SEM_R | SEM_A)))
    {
      perror("Could not create semaphore\n");
      exit(1);
    }
  
  
  /*Get the shared memory*/
  if (-1==(shmid=shmget(IPC_PRIVATE, sizeof(life_state),SEM_R | SEM_A)))
    {
      perror("Could not shmget in life3d.c");
      exit(1);
    }
   
  /* Attach to shared memory */
  if ((life_state *)-1==(pLifeState=(life_state *)shmat(shmid, 0, 0)))
    {
      perror("Could not shmat in life3d.c");
      exit(1);
    }
  

}

/* =================================================== */
/* ================= CleanUpIPC() ==================== */
void CleanUpIPC()
{
  /* performs cleanup, returning any IPC resources to the system */
  /* Detach and clear shared memory */
  if (-1==shmdt((void*)pLifeState))
    {
      perror("Could not shmdt in CleanUpIPC in life3d.c");
      exit(1);
    }
  if (-1==shmctl(shmid, IPC_RMID, NULL))
    {
      perror("Could not shmctl in CleanUpIPC in life3d.c");
        exit(1);
    }
  /* remove semaphores*/
  sem_rm(sem_cleanup_watchdog);
  sem_rm(sem_shared_mem);
  sem_rm(sem_time_slice_start);
  sem_rm(sem_time_slice_end);
  sem_rm(sem_num_updated_faces);
  sem_rm(sem_faces_commited_next_state);
  
}


/* =================================================== */
/* ============== dbg_printsem() ===================== */
void dbg_printsem(int sem)
{
  /*used only in debugging, prints semaphore's value */
  int sval;
  if ((sval=semctl(sem,0,GETVAL,0))<0)
    perror("Can't getval\n");
#ifdef DEBUG_MODE
  printf("sem val is %d\n",sval);
#endif
}



/* =================================================== */
/* ================= start_watchdog() ================ */
/* fork off a watchdog that will cleanup the IPC resources
   when all the processes using them exit*/

void fork_watchdog()
{
  int fork_retval; /*used when forking a watchdog*/
  
  /*fork of a watchdog that will cleanup all IPC resources when all
    the processes in the game have exited*/
  switch (fork_retval=fork())
    {
    case -1:
      perror("Error: could not fork a watchdog process!\n");
      exit(1);
      break;
    case 0:
      /*in the watchdog*/

      /*create a new session*/ 
      if(-1==setsid()){
	perror("Could not create a new session in watchdog\n");
	exit(1);
      }    
      /*Just wait on the semaphore that drops to 0 when all the processes
	have exited, and then cleanup the IPC resources */
      sem_op(sem_cleanup_watchdog,0);
      printf("watchdog: cleanup up IPC's\n");
      /*cleanup ipc resources*/
      CleanUpIPC();
      exit(0);
      /*never reached*/
      break;
    }
}

/* =================================================== */
/* =================fork_face_workers() ============== */
/* Forks off the faces
*/

void fork_face_workers()
{
  int i;
  int fork_retval;
  
  
  for(i=0;i<FACES;i++)
    {
      switch (fork_retval=fork())
        {
        case -1:
          perror("Error: could not fork a face process!\n");
          exit(1);
          break;
        case 0:
          /*We're now in the child*/
          semsignalundo(sem_cleanup_watchdog);
          
          RunLifeOnFace(i);
          /*This point should never be reached*/
          exit(1);
          break;
        }
    }
}


/* =================================================== */
/* ======================= main() ==================== */
/* initializes the player
   and then runs the main player simulation loop

   will never exit (must be killed with CTRL-C)
*/
int main(int argc,char** argv)
{
  int i,j,k;
  int test_enabled = argc >1 && !strcmp(argv[1],"-test");
  printf("Running life 3D%s Ctrl-C to exit\n",
         test_enabled ? " with test pattern." : "with random patterns.");
  srand(getpid());
  
  /*Initialize IPC resources*/
  InitLifeIPCResources();

  fork_watchdog();

  fork_face_workers();

  LVLibInit(argc,argv);
  LVLibStartDisplay();
  /*Enable auto rotation (spinning cube)
    set this to LVLFalse to disable automatic spinning
    then use the LVLSet?angle() calls to set whatever
    viewing angles you want. Do this before calling
    LVLibUpdateDisplay*/
  LVLibSetAutoRotate(LVLTrue);


  semwait(sem_shared_mem);
  for(k=0;k<FACES;k++)
    {
      for(i=0;i<N;i++)
        for(j=0;j<N;j++)
          pLifeState->cells[k][i][j]=test_enabled ?
            CELL_DEAD : rand()%2;
      if(test_enabled)
        {
          pLifeState->cells[k][1][0]=CELL_ALIVE;
          pLifeState->cells[k][2][0]=CELL_ALIVE;
          pLifeState->cells[k][1][1]=CELL_ALIVE;
          pLifeState->cells[k][1][2]=CELL_ALIVE;
          pLifeState->cells[k][0][1]=CELL_ALIVE;
        }
    }
  semsignal(sem_shared_mem);

  
  /* Start the main life loop */
  for(;;){
    /*Do the display*/
    semwait(sem_shared_mem);
    for(i=0;i<FACES;i++)
      LVLibSetFaceCells(i,pLifeState->cells);
    LVLibUpdateDisplay();
    semsignal(sem_shared_mem);
    
    
    /*raise sem to +Num faces to start the time slice*/
    sem_op(sem_time_slice_start,FACES);
    
    /*wait for all the faces to finish updating their next states, 
      this is important because it ensures that players have finished
      their calculations and moved to a new position, so it is safe to 
      change the current position of the ball as no player
      will be looking  at it at this stage.
    */
    
    sem_op(sem_num_updated_faces,-FACES);
    /*At this point all faces have computed their next states*/
    

    /*raise sem to indicate to the faces the slice is over*/
    sem_op(sem_time_slice_end,FACES);

    /*Faces should commit next state to the current state here*/
    sem_op(sem_faces_commited_next_state,-FACES);
    
    /*make sure all the players are done with the time slice
      before proceeding*/
    sem_op(sem_time_slice_end,0);
    
  }
  return 0;
}
